"""Starlette helpers for WTForms

Contains code with and without modification from Flask-WTF:

Copyright (c) 2010 by Dan Jacob.
Copyright (c) 2013 by Hsiaoming Yang.

Some rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.

* The names of the contributors may not be used to endorse or
  promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from starlette.datastructures import ImmutableMultiDict
from starlette.requests import Request as StarletteRequest
from wtforms import Form
from wtforms.meta import DefaultMeta

from starlette_wtf.util import get_formdata


__all__ = ['StarletteForm']


SUBMIT_METHODS = set(('POST', 'PUT', 'PATCH', 'DELETE'))
_Auto = object()


class StarletteForm(Form):
    """Starlette-specific subclass of WTForms :class:`~wtforms.form.Form`.

    To populate from submitted formdata use the ```.from_submit()``` class
    method to initialize the instance.
    """

    def __init__(self, request: StarletteRequest, *args, **kwargs):
        """Initialize StarletteForm instance.

        Args:
          request (:class:`starlette.requests.Request`): The request instance.
          *args: Variable length argument list for :class:`starlette.requests
              .Request`
          **kwargs: Arbitrary keyword arguments for :class:`starlette.requests
              .Request`

        """
        # cache request
        self._request = request
        
        # for WTForms CSRF handling
        if hasattr(request.state, 'csrf_config'):
            config = request.state.csrf_config
            kwargs['meta'] = {
                'csrf': config['enabled'],
                'csrf_secret': str(config['csrf_secret']).encode('utf-8'),
                'csrf_class': config['csrf_class'],
                'csrf_context': request,
                'csrf_field_name': config['csrf_field_name'],
                'csrf_time_limit': config['csrf_time_limit']
            }

        super().__init__(*args, **kwargs)

        
    @classmethod
    async def from_formdata(cls, request: StarletteRequest, formdata=_Auto,
                            **kwargs):
        """Method to support initializing class from submitted formdata. If
        request is a POST, PUT, PATCH or DELETE, form will be initialized using
        formdata. Otherwise, it will be initialized using defaults.

        Args:
          request (:class:`starlette.requests.Request`): The request instance.
          formdata (ImmutableMultiDict, optional): If present, this will be
              used to initialize the form fields.

        Returns:
          :class:`starlette_wtf.form.StarletteForm`: A new form instance.

        """
        if formdata is _Auto:
            if request.method in SUBMIT_METHODS:
                # get formdata from request.form() or request.json()
                formdata = await get_formdata(request)
            else:
                formdata = None
            
        # return new instance
        return cls(request, formdata=formdata, **kwargs)
    
    
    def is_submitted(self):
        """Consider the form submitted if there is an active request and
        the method is ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
        """
        return self._request.method in SUBMIT_METHODS
    
        
    def validate_on_submit(self):
        """Call :meth:`validate` only if the form is submitted.
        This is a shortcut for ``form.is_submitted() and form.validate()``.
        """
        return self.is_submitted() and self.validate()



"""Starlette helpers for WTForms

Contains code with and without modification from Flask-WTF:

Copyright (c) 2010 by Dan Jacob.
Copyright (c) 2013 by Hsiaoming Yang.

Some rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.

* The names of the contributors may not be used to endorse or
  promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import hashlib
import hmac
import os

from itsdangerous import BadData, SignatureExpired, URLSafeTimedSerializer
from starlette.datastructures import ImmutableMultiDict
from starlette.requests import Request as StarletteRequest
from wtforms import ValidationError


async def get_formdata(request: StarletteRequest):
    """Return formdata from request. Handles multi-dict and json content types.

    Args:
      request (:class:`starlette.requests.Request`): The request instance.

    Returns:
      formdata (ImmutableMultiDict): The form/json data from the request.

    """
    if request.headers.get('content-type') == 'application/json':
        formdata = ImmutableMultiDict(await request.json())
    else:
        formdata = await request.form()

    return formdata


def generate_csrf(request: StarletteRequest,
                  secret_key: str,
                  field_name: str):
    """Generate a new token, store it in the session and return a time-signed
    token. If a token is already present in the session, it will be used to
    generate a new time signed token. The time-signed token is cached per
    request so multiple calls to this function will return the same time-signed
    token.

    Args:
      request (:class:`starlette.requests.Request`): The request instance.
      secret_key (str): The signing key.
      field_name (str): Where the token is stored in the session.

    Returns:
      str: The time-signed token

    """
    if not hasattr(request.state, field_name):
        s = URLSafeTimedSerializer(secret_key, salt='wtf-csrf-token')

        session = request.session

        # get/set token in session
        if field_name not in session:
            session[field_name] = hashlib.sha1(os.urandom(64)).hexdigest()

        try:
            token = s.dumps(session[field_name])
        except TypeError:
            session[field_name] = hashlib.sha1(os.urandom(64)).hexdigest()
            token = s.dumps(session[field_name])

        setattr(request.state, field_name, token)

    return getattr(request.state, field_name)


def validate_csrf(request: StarletteRequest,
                  data: str,
                  secret_key: str,
                  field_name: str,
                  time_limit: int=None):
    """Check if the given data is a valid CSRF token. This compares the given
    signed token to the one stored in the session.

    Args:
      request (:class:`starlette.requests.Request`): The request instance.
      data (str): The time-signed CSRF token to be checked.
      secret_key (str): The signing key Used to sign the token.
      field_name (str): Where token is stored in session.
      time_limit (int): Number of seconds that the token is valid. Defaults to
          None.

    Returns:
      None: Completes successfully otherwise raises an error.

    Raises:
      ValidationError: If token failes validation. Contains the reason that
          validation failed.

    """
    if not data:
        raise ValidationError('The CSRF token is missing.')

    if field_name not in request.session:
        raise ValidationError('The CSRF session token is missing.')

    s = URLSafeTimedSerializer(secret_key, salt='wtf-csrf-token')
    
    try:
        token = s.loads(data, max_age=time_limit)
    except SignatureExpired:
        raise ValidationError('The CSRF token has expired.')
    except BadData:
        raise ValidationError('The CSRF token is invalid.')

    # do safe string comparison
    if not hmac.compare_digest(request.session[field_name], token):
        raise ValidationError('The CSRF tokens do not match.')



from starlette_wtf.csrf import (CSRFProtectMiddleware, csrf_protect,
                                csrf_token, CSRFError)
from starlette_wtf.form import StarletteForm


__all__ = [
    'StarletteForm',
    'CSRFProtectMiddleware',
    'csrf_protect',
    'csrf_token',
    'CSRFError'
    ]



#!/usr/bin/env python

from setuptools import setup, find_packages


def get_long_description():
    with open('README.md') as f:
        return f.read()
    

setup(
    name="Starlette-WTF",
    python_requires=">=3.6",
    version="0.1",
    url="https://github.com/muicss/starlette-wtf",
    license="MIT",
    author="Andres Morey",
    description="Simple integration of Starlette and WTForms.",
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    packages=['starlette_wtf'],
    platforms='any',
    install_requires=[
        'itsdangerous',
        'python-multipart',
        'starlette',
        'WTForms'
    ],
    extras_require={
        'test': ['pytest', 'requests']
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    zip_safe=False
    )



"""Starlette helpers for WTForms

Contains code with and without modification from Flask-WTF:

Copyright (c) 2010 by Dan Jacob.
Copyright (c) 2013 by Hsiaoming Yang.

Some rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials provided
  with the distribution.

* The names of the contributors may not be used to endorse or
  promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from typing import List, Optional, ByteString
import functools
from urllib.parse import urlparse

from starlette.applications import Starlette as StarletteApplication
from starlette.exceptions import HTTPException
from starlette.middleware.base import BaseHTTPMiddleware
from wtforms import ValidationError
from wtforms.csrf.core import CSRF

from starlette_wtf.form import SUBMIT_METHODS, get_formdata
from starlette_wtf.util import generate_csrf, validate_csrf


__all__ = ['CSRFProtectMiddleware', 'csrf_protect', 'csrf_token', 'CSRFError']


class _StarletteFormCSRF(CSRF):
    def setup_form(self, form):
        self.form_meta = form.meta
        return super(_StarletteFormCSRF, self).setup_form(form)

    
    def generate_csrf_token(self, csrf_token_field):
        meta = self.form_meta
        return generate_csrf(
            request=meta.csrf_context,
            secret_key=meta.csrf_secret,
            field_name=meta.csrf_field_name)

    
    def validate_csrf_token(self, form, field):
        meta = self.form_meta

        if hasattr(meta.csrf_context.state, 'csrf_valid'):
            # already validated by CSRFProtectMiddleware
            return

        validate_csrf(
            request=meta.csrf_context,
            data=field.data,
            secret_key=meta.csrf_secret,
            field_name=meta.csrf_field_name,
            time_limit=meta.csrf_time_limit)


DEFAULT_ENABLED = True
DEFAULT_CSRF_SECRET = None
DEFAULT_CSRF_FIELD_NAME = 'csrf_token'
DEFAULT_CSRF_TIME_LIMIT = 3600
DEFAULT_CSRF_HEADERS = ['X-CSRFToken', 'X-CSRF-Token']
DEFAULT_CSRF_SSL_STRICT = True


class CSRFProtectMiddleware(BaseHTTPMiddleware):
    def __init__(self,
                 app: StarletteApplication,
                 enabled: bool=DEFAULT_ENABLED,
                 csrf_secret: Optional[ByteString]=DEFAULT_CSRF_SECRET,
                 csrf_field_name: str=DEFAULT_CSRF_FIELD_NAME,
                 csrf_time_limit: int=DEFAULT_CSRF_TIME_LIMIT,
                 csrf_headers: List[str]=DEFAULT_CSRF_HEADERS,
                 csrf_ssl_strict: bool=DEFAULT_CSRF_SSL_STRICT):
        """ASGI Middleware needed by Starlette-WTF to enable CSRF protection.
        
        Args:
          app (:class:`starlette.applications.Starlette`): The application
              instance.
          enabled (bool, optional): Defaults to True.
          csrf_secret (str): CSRF secret key
          csrf_field_name (str): The CSRF token field name. Defaults to
              "csrf_token".
          csrf_time_limit (int): The CSRF token time limit in seconds. Defaults
              to 3600.
          csrf_headers (list of str): CSRF HTTP header field names. Defaults to
              ["X-CSRFToken", "X-CSRF-Token"]
          csrf_ssl_strict (bool): If enabled, ensures same origin policy on
              https requests. Defaults to true.

        """
        if enabled and not csrf_secret:
            raise RuntimeError('`csrf_secret` is required')

        self.csrf_config = {
            'enabled': enabled,
            'csrf_secret': csrf_secret,
            'csrf_class': _StarletteFormCSRF,
            'csrf_field_name': csrf_field_name,
            'csrf_time_limit': csrf_time_limit,
            'csrf_headers': csrf_headers,
            'csrf_ssl_strict': csrf_ssl_strict
        }
        
        super().__init__(app)

        
    async def dispatch(self, request, call_next):
        """Add CSRF config to request state
        """
        request.state.csrf_config = self.csrf_config
        return await call_next(request)


class CSRFError(HTTPException):
    """HTTPException class for CSRF validation errors. Returns 403 Forbidden.
    """
    def __init__(self, detail=None):
        super().__init__(status_code=403, detail=detail)


def csrf_protect(func):
    """Returns decorator that performs CSRF validation before calling 
    endpoint function
    """
    @functools.wraps(func)
    async def endpoint_wrapper(request, *args, **kwargs):
        if not request.method in SUBMIT_METHODS:
            return await func(request, *args, **kwargs)
        
        # get token
        signed_token = await get_csrf_token(request)

        config = request.state.csrf_config
        
        # validate token
        try:
            validate_csrf(request,
                          signed_token,
                          secret_key=config['csrf_secret'],
                          field_name=config['csrf_field_name'],
                          time_limit=config['csrf_time_limit'])
        except ValidationError as e:
            raise CSRFError(e.args[0])

        # strict ssl check
        if request.url.scheme == 'https' and config['csrf_ssl_strict']:
            referrer = request.headers.get('REFERER')

            if not referrer:
                raise CSRFError('The referrer header is missing.')

            if not same_origin(urlparse(referrer), request.url):
                raise CSRFError('The referrer does not match the host.')
        
        # mark request as valid
        request.state.csrf_valid = True

        # pass on request
        return await func(request, *args, **kwargs)

    return endpoint_wrapper


def csrf_token(request):
    """Return CSRF token

    Args:
      request (:class:`starlette.requests.Request`): The request instance.

    Returns:
      str: The signed token

    """
    csrf_config = request.state.csrf_config
    
    return generate_csrf(request,
                         secret_key=csrf_config['csrf_secret'],
                         field_name=csrf_config['csrf_field_name'])


async def get_csrf_token(request):
    csrf_config = request.state.csrf_config
    
    formdata = await get_formdata(request)
    
    field_name = csrf_config['csrf_field_name']
    base_token = formdata.get(field_name)

    if base_token:
        return base_token
    
    # find the token in the headers
    for header_name in csrf_config['csrf_headers']:
        csrf_token = request.headers.get(header_name)
        if csrf_token:
            return csrf_token
        
    return None


def same_origin(url1, url2):
    return (
        url1.scheme == url2.scheme
        and url1.hostname == url2.hostname
        and url1.port == url2.port
    )
